% The MATLAB code of training algorithm and testing results for the NARX DCFS
% This code has been taken and adapted for NARX DCFS from the reference:
%
% Wang, L.-X. "Fast Training Algorithms for Deep Convolutional Fuzzy Systems with Application to Stock Index Prediction". 
% IEEE Trans. Fuzzy Syst. 2019, 1–1, doi:10.1109/TFUZZ.2019.2930488. 
%
% DCFS model for examples presented in chapter 4.1. Gas Furnace Model Identification 
% An example of General NARX DCFS - Definition 1
clear all

% Choice of the model structure
%   1 - Wang Mendel Fuzzy system vith 6 inputs
%   2 - General_NARX_DCFS
%   3 - Input-output NARX DCFS
%   4 - Sub-model NARX DCFS
model=1; 

% The number of fuzzy sets covering inputs
Num_Fuzzy_Sets=30; 

% Read Box Jenkin's data set from file
fid = fopen('boxjen.dat','r'); 
[data]=fscanf(fid,'%f',[3,296]);
fclose(fid);

u=data(2,:)'; y_p=data(3,:)'; % Input and output if the process

nu=3;ny=3;d=3; % Regression vector order and delay

l=max(nu+d,ny);
% The following lines generate data_matrix, which is N×(nu+ny+1) matrix containing all the data, the first
% 6 columns are inputs u(k-d-1),...,u(k-d-ny),y(k-1),...,y(k-ny) and the last column is output y(k).
for i=1:size(y_p,1)-l
  k=1;  
  for j=l-d:-1:l-d-nu+1
     data_matrix(i,k)=u(i+j-1); 
     k=k+1;
  end
  k=1;
  for j=l:-1:l-ny+1
     data_matrix(i,nu+k)=y_p(i+j-1); 
     k=k+1;
  end
  data_matrix(i,nu+ny+1)=y_p(i+l);
end 
% data_matrix is matrix containing all the data, the first
% 11 columns are inputs and the last column is output.
[N,~]=size(data_matrix);
ntrain=round(N*200/290); % The first 5/6 are training data, and the
% remaining 1/6 are testing data.

% Strat training
t=cputime;


if model == 1  %    WMFS
    % Testing different NARX DCFS structures
    % Wamg Mendel Fuzzy System
    Num_Fuzzy_Sets=11;
    for i=1:ntrain
      for j=1:3
        x11(i,j)=data_matrix(i,j); x11(i,3+j)=data_matrix(i,nu+j);  
      end 
      y(i,1)=data_matrix(i,7); % y is the desired output.
    end;
    [y_center_11 MinMaxRange_11]=train_wm_deep(Num_Fuzzy_Sets,x11,y);
    % Training done
    for i=1:N
      for j=1:3
        x11(i,j)=data_matrix(i,j); x11(i,3+j)=data_matrix(i,nu+j);  
      end 
      y(i,1)=data_matrix(i,7); % y is the desired output.
    end;
    % Level 1 computing done
    y_model=eval_wm_deep(Num_Fuzzy_Sets,y_center_11,MinMaxRange_11,x11); % yy is output of DCFS.
end    

if model == 2   %      General NARX DCFS
    % General Structure of a NARX DCFS. 
    % An example of general NARX DCFS with 
    % nu = 3, ny =3, n1 =4, m^1=m^2=3,m^3=2 and L =3 which is presented
    % in Figure 4a in paper.
    % The input variables of the 4 FS's in first Level with moving window 3.
    for i=1:ntrain
      for j=1:3
        x11(i,j)=data_matrix(i,j);
        x12(i,j)=data_matrix(i,j+1);
        x13(i,j)=data_matrix(i,j+2);
        x14(i,j)=data_matrix(i,j+3);
      end 
      y(i,1)=data_matrix(i,7); % y is the desired output.
    end;
    
    % The next 4 lines train the 4 FS's in Level 1, where
    % train_wm_deep trains the fuzzy system.
    [y_center_11 MinMaxRange_11]=train_wm_deep(Num_Fuzzy_Sets,x11,y);
    [y_center_12 MinMaxRange_12]=train_wm_deep(Num_Fuzzy_Sets,x12,y);
    [y_center_13 MinMaxRange_13]=train_wm_deep(Num_Fuzzy_Sets,x13,y);
    [y_center_14 MinMaxRange_14]=train_wm_deep(Num_Fuzzy_Sets,x14,y);
    % The following 6 lines computes x21,…,x22 which are inputs to the 2 fuzzy
    % systems in Level 2, where eval_wm_deep computes the output of the
    % fuzzy system with parameters y_center and MinMaxRange.
    x21(1:ntrain,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_11,MinMaxRange_11,x11);
    x21(1:ntrain,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_12,MinMaxRange_12,x12);
    x21(1:ntrain,3)=eval_wm_deep(Num_Fuzzy_Sets,y_center_13,MinMaxRange_13,x13);
    x22(1:ntrain,1)=x21(:,2);
    x22(1:ntrain,2)=x21(:,3);
    x22(1:ntrain,3)=eval_wm_deep(Num_Fuzzy_Sets,y_center_14,MinMaxRange_14,x14);
    % The following 2 lines train the 2 fuzzy systems in Level 2.
    [y_center_21 MinMaxRange_21]=train_wm_deep(Num_Fuzzy_Sets,x21,y);
    [y_center_22 MinMaxRange_22]=train_wm_deep(Num_Fuzzy_Sets,x22,y);
    % % x31 are inputs to the fuzzy system in Level 3.
    x31(1:ntrain,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_21,MinMaxRange_21,x21);
    x31(1:ntrain,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_22,MinMaxRange_22,x22);
    % Train the fuzzy system in Level 3.
    [y_center_31 MinMaxRange_31]=train_wm_deep(Num_Fuzzy_Sets,x31,y);
    
    % Computing the training and testing errors 
    % The variables are the same.
    for i=1:N
      for j=1:3
        x11(i,j)=data_matrix(i,j);
        x12(i,j)=data_matrix(i,j+1);
        x13(i,j)=data_matrix(i,j+2);
        x14(i,j)=data_matrix(i,j+3);
      end 
      y(i,1)=data_matrix(i,7); % y is the desired output.
    end;
    
    
    x21(1:N,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_11,MinMaxRange_11,x11);
    x21(1:N,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_12,MinMaxRange_12,x12);
    x21(1:N,3)=eval_wm_deep(Num_Fuzzy_Sets,y_center_13,MinMaxRange_13,x13);
    x22(1:N,1)=x21(:,2);
    x22(1:N,2)=x21(:,3);
    x22(1:N,3)=eval_wm_deep(Num_Fuzzy_Sets,y_center_14,MinMaxRange_14,x14);
    % Level 1 
    x31(1:N,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_21,MinMaxRange_21,x21);
    x31(1:N,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_22,MinMaxRange_22,x22);
    % Level 2 
    y_model=eval_wm_deep(Num_Fuzzy_Sets,y_center_31,MinMaxRange_31,x31); % yy is output of DCFS.
    % Level 3 c
end

if model == 3  %    Input-output NARX DCFS
    % General Structure of a NARX DCFS. 
    % in Figure 4b in paper.
    % The input variables of the 3 FS's in first Level with moving window 3.
    for i=1:ntrain
      for j=1:2
        x11(i,j)=data_matrix(i,j);
        x12(i,j)=data_matrix(i,j+1);
        x13(i,j)=data_matrix(i,j+3);
        x14(i,j)=data_matrix(i,j+4);
      end 
      y(i,1)=data_matrix(i,7); % y is the desired output.
    end;
    
    % The next 4 lines train the 4 FS's in Level 1, where
    % train_wm_deep trains the fuzzy system.
    [y_center_11 MinMaxRange_11]=train_wm_deep(Num_Fuzzy_Sets,x11,y);
    [y_center_12 MinMaxRange_12]=train_wm_deep(Num_Fuzzy_Sets,x12,y);
    [y_center_13 MinMaxRange_13]=train_wm_deep(Num_Fuzzy_Sets,x13,y);
    [y_center_14 MinMaxRange_14]=train_wm_deep(Num_Fuzzy_Sets,x14,y);

    % The following 4 lines computes x21,…,x22 which are inputs to the 2 fuzzy
    % systems in Level 2, where eval_wm_deep computes the output of the
    % fuzzy system with parameters y_center and MinMaxRange.
    x21(1:ntrain,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_11,MinMaxRange_11,x11);
    x21(1:ntrain,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_12,MinMaxRange_12,x12);
    x21(1:ntrain,3)=eval_wm_deep(Num_Fuzzy_Sets,y_center_13,MinMaxRange_13,x13);
    x22(1:ntrain,1)=x21(:,2);
    x22(1:ntrain,2)=x21(:,3);
    x22(1:ntrain,3)=eval_wm_deep(Num_Fuzzy_Sets,y_center_14,MinMaxRange_14,x14);

    % The following 2 lines train the 2 fuzzy systems in Level 2.
    [y_center_21 MinMaxRange_21]=train_wm_deep(Num_Fuzzy_Sets,x21,y);
    [y_center_22 MinMaxRange_22]=train_wm_deep(Num_Fuzzy_Sets,x22,y);
    % x31 are inputs to the fuzzy system in Level 3.
    x31(1:ntrain,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_21,MinMaxRange_21,x21);
    x31(1:ntrain,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_22,MinMaxRange_22,x22);
    % Train the fuzzy system in Level 3.
    [y_center_31 MinMaxRange_31]=train_wm_deep(Num_Fuzzy_Sets,x31,y);
    
    % Computing the training and testing errors 
    % The variables are the same.
    % The input variables of the 3 FS's in first Level with moving window 3.
    for i=1:N
      for j=1:2
        x11(i,j)=data_matrix(i,j);
        x12(i,j)=data_matrix(i,j+1);
        x13(i,j)=data_matrix(i,j+3);
        x14(i,j)=data_matrix(i,j+4);
      end 
      y(i,1)=data_matrix(i,7); % y is the desired output.
    end;
    
    x21(1:N,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_11,MinMaxRange_11,x11);
    x21(1:N,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_12,MinMaxRange_12,x12);
    x21(1:N,3)=eval_wm_deep(Num_Fuzzy_Sets,y_center_13,MinMaxRange_13,x13);
    x22(1:N,1)=x21(:,2);
    x22(1:N,2)=x21(:,3);
    x22(1:N,3)=eval_wm_deep(Num_Fuzzy_Sets,y_center_14,MinMaxRange_14,x14);
    % Level 1 done
    x31(1:N,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_21,MinMaxRange_21,x21);
    x31(1:N,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_22,MinMaxRange_22,x22);
    % Level 2 done  
    y_model=eval_wm_deep(Num_Fuzzy_Sets,y_center_31,MinMaxRange_31,x31); % yy is output of DCFS.
    % Level 3 done
end

if model == 4  %    Sub-model NARX DCFS
    % in Figure 4c in paper.    
    % The input variables of the 3 FS's in first Level with moving window 2.
    for i=1:ntrain
        x11(i,1)=data_matrix(i,1);
        x11(i,2)=data_matrix(i,4);
        x12(i,1)=data_matrix(i,2);
        x12(i,2)=data_matrix(i,5);
        x13(i,1)=data_matrix(i,3);
        x13(i,2)=data_matrix(i,6);
        y(i,1)=data_matrix(i,7); % y is the desired output.
    end;
    
    % The next 3 lines train the 3 FS's in Level 1, where
    % train_wm_deep trains the fuzzy system.
    [y_center_11 MinMaxRange_11]=train_wm_deep(Num_Fuzzy_Sets,x11,y);
    [y_center_12 MinMaxRange_12]=train_wm_deep(Num_Fuzzy_Sets,x12,y);
    [y_center_13 MinMaxRange_13]=train_wm_deep(Num_Fuzzy_Sets,x13,y);

    % The following 4 lines computes x21,…,x22 which are inputs to the 2 fuzzy
    % systems in Level 2, where eval_wm_deep computes the output of the
    % fuzzy system with parameters y_center and MinMaxRange.
    x21(1:ntrain,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_11,MinMaxRange_11,x11);
    x21(1:ntrain,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_12,MinMaxRange_12,x12);
    x22(1:ntrain,1)=x21(1:ntrain,2);
    x22(1:ntrain,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_13,MinMaxRange_13,x13);

    % The following 2 lines train the 2 fuzzy systems in Level 2.
    [y_center_21 MinMaxRange_21]=train_wm_deep(Num_Fuzzy_Sets,x21,y);
    [y_center_22 MinMaxRange_22]=train_wm_deep(Num_Fuzzy_Sets,x22,y);
    % x31 are inputs to the fuzzy system in Level 3.
    x31(1:ntrain,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_21,MinMaxRange_21,x21);
    x31(1:ntrain,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_22,MinMaxRange_22,x22);
    % Train the fuzzy system in Level 3.
    [y_center_31 MinMaxRange_31]=train_wm_deep(Num_Fuzzy_Sets,x31,y);
    
    % Computing the training and testing errors 
    % The variables are the same.
    for i=1:N
        x11(i,1)=data_matrix(i,1);
        x11(i,2)=data_matrix(i,4);
        x12(i,1)=data_matrix(i,2);
        x12(i,2)=data_matrix(i,5);
        x13(i,1)=data_matrix(i,3);
        x13(i,2)=data_matrix(i,6);
        y(i,1)=data_matrix(i,7); % y is the desired output.
    end;
    x21(1:N,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_11,MinMaxRange_11,x11);
    x21(1:N,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_12,MinMaxRange_12,x12);
    x22(1:N,1)=x21(:,2);
    x22(1:N,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_13,MinMaxRange_13,x13);   
    % Level 1 done
    x21(1:N,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_11,MinMaxRange_11,x11);
    x21(1:N,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_12,MinMaxRange_12,x12);
    x22(1:N,1)=x21(:,2);
    x22(1:N,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_13,MinMaxRange_13,x13);
    % Level 2 done
    x31(1:N,1)=eval_wm_deep(Num_Fuzzy_Sets,y_center_21,MinMaxRange_21,x21);
    x31(1:N,2)=eval_wm_deep(Num_Fuzzy_Sets,y_center_22,MinMaxRange_22,x22);
    
    y_model=eval_wm_deep(Num_Fuzzy_Sets,y_center_31,MinMaxRange_31,x31); % yy is output of DCFS.
    % Level 3 done
end

y_model=y_model';

t_cpu=cputime-t

% The following computes and plots the training and testing errors.
for i=1:N
 e(i)=y(i)-y_model(i);
end;
err_train=0;
for i=1:ntrain
 err_train=err_train+e(i)^2;
end;
MSE_train=err_train/ntrain;
RMSE_train=sqrt(err_train/ntrain);
err_test=0;
for i=ntrain+1:N
 err_test=err_test+e(i)^2;
end;
MSE_test=err_test/(N-ntrain);
RMSE_test=sqrt(err_test/(N-ntrain));

% Figure
t=0:9:(N-1)*9; t=t';
fig1=figure(1);
fig1.Position=[10 10 1120 840*3/4];

ax1=subplot(3,1,1), p1=plot(t,x11(:,1),'b'); ax1.FontSize = 11; %axes('Position',[0.13,0.7101,0.775,0.1953] )
axis([0 max(t) min(u)-abs((max(u)-min(u)))*0.05 max(u)+abs((max(u)-min(u)))*0.05]), grid on, grid minor,p1.LineWidth=2;
ylabel('\itu\rm','FontSize',12), title('\rmInput - Gas Rate','FontSize',12), 
line([ntrain*9 ntrain*9],[min(u)-abs((max(u)-min(u)))*0.05 max(u)+abs((max(u)-min(u)))*0.05],'Color','Green','LineStyle','--','LineWidth',2)
legend(p1,'\itu\rm','FontSize',12,'TextColor','black')
text(1600,max(u)-(max(u)-min(u))*0.1,['Train'],'FontSize',12), text(1900,max(u)-(max(u)-min(u))*0.1,['Test'],'FontSize',12)

ax2=subplot(3,1,2), p2=plot(t,y,t,y_model); ax2.FontSize = 11;
title('\rmOutput - CO2 concentration','FontSize',12), ylabel('\ity, y_m','FontSize',12), axis([0 max(t) min(y)-abs((max(y)-min(y)))*0.05 max(y)+abs((max(y)-min(y)))*0.05]), grid on, grid minor, p2(1).LineWidth=2; p2(2).LineWidth=2;
line([ntrain*9 ntrain*9],[min(y)-abs((max(y)-min(y)))*0.05 max(y)+abs((max(y)-min(y)))*0.05],'Color','Green','LineStyle','--','LineWidth',2)
legend(p2,'\ity\rm','\ity_m','FontSize',12,'TextColor','black','Location','best')

ax3=subplot(3,1,3), p3=plot(t,e,'r'); ax3.FontSize = 11;
title(['\rm\it\tau\rm = ' num2str(d) ',     \rm\itM\rm = ' num2str(Num_Fuzzy_Sets) ',     \rm\itMSE_{train}\rm = ' num2str(MSE_train) ',     \rm\itMSE_{test}\rm = ' num2str(MSE_test) ',     \rm\itt_{CPU}\rm = ' num2str(t_cpu) ' \rms'],'FontSize',12) 
xlabel('\itt \rm[s]','FontSize',12), ylabel('\ite','FontSize',12), axis([0 max(t) min(e)-(max(e)-min(e))*0.05 max(e)+(max(e)-min(e))*0.05]), grid on, grid minor, p3.LineWidth=2;
line([ntrain*9 ntrain*9],[min(e)-abs((max(e)-min(e)))*0.05 max(e)+abs((max(e)-min(e)))*0.05],'Color','Green','LineStyle','--','LineWidth',2)
legend(p3,'\ite\rm','FontSize',12,'TextColor','black','Location','best')

function [y_center ranges]=train_wm_deep(Num_Fuzzy_Sets,data_matrix,y)
% Train fuzzy system with input data_matrix and output y to get y_center and ranges,
% where y_center is the in WM FS and ranges is the endpoints in.
% Original function is here:  https://ieeexplore.ieee.org/ielx7/91/9130783/8788632/code_training_algorithm.pdf?arnumber=8788632
%
extra=[data_matrix,y];
[numSamples,m]=size(extra);
numInput=m-1;
for i=1:numInput
  fnCounts(i)=Num_Fuzzy_Sets;
end;
ranges = zeros(numInput,2);
activFns = zeros(numInput,2);
activGrades = zeros(numInput,2);
searchPath = zeros(numInput,2);
numCells = 1; % number of regions (cells)
for i = 1:numInput
  ranges(i,1) = min(extra(:,i));
  ranges(i,2) = max(extra(:,i));
  numCells = numCells * fnCounts(i);
end;
baseCount(1)=1;
for i=2:numInput
  baseCount(i)=1;
for j=2:i
  baseCount(i)=baseCount(i)*fnCounts(numInput-j+2);
end;
end;
% Generate rules for cells covered by data
y_center = zeros(1,numCells); % THEN part centers of generated rules
ym = zeros(1,numCells);
for k = 1:numSamples
  for i = 1:numInput
    numFns = fnCounts(i);
    nthActive = 1;
    for nthFn = 1:numFns
      grade = Membership_at_x(numFns,nthFn,extra(k,i),ranges(i,1),ranges(i,2));
      if grade > 0
        activFns(i,nthActive) = nthFn;
        activGrades(i,nthActive) = grade;
        nthActive = nthActive + 1;
      end;
    end; % endfor nthFn
  end; % endfor i
  for i=1:numInput
    if activGrades(i,1) >= activGrades(i,2)
      searchPath(i,1)=activFns(i,1);
      searchPath(i,2)=activGrades(i,1);
    else
      searchPath(i,1)=activFns(i,2);
      searchPath(i,2)=activGrades(i,2);
    end;
  end;
  indexcell=1;
  grade=1;
  for i=1:numInput
    grade=grade*searchPath(i,2);
    indexcell=indexcell+(searchPath(numInput-i+1,1)-1)*baseCount(i);
  end;
  ym(indexcell)=ym(indexcell)+grade;
  y_center(indexcell)=y_center(indexcell)+extra(k,numInput+1)*grade;
end; % endfor k
for j=1:numCells
  if ym(j) ~= 0
    y_center(j)=y_center(j)/ym(j);
  end;
end;
% Extrapolate the rules to all the cells
for i=1:numInput-1
  baseCount(i)=1;
  for j=1:i
    baseCount(i)=baseCount(i)*fnCounts(numInput-j+1);
  end;
end;
ct=1;
y_center_b = zeros(1,numCells);
yNum_Fuzzy_Sets = zeros(1,numCells);
while ct > 0
  ct=0;
  for s=1:numCells
    if ym(s) == 0
      ct=ct+1;
    end;
  end;
  for s=1:numCells
    if ym(s) == 0
      s1=s;
      index = ones(1,numInput);
      for i=numInput-1:-1:1
        while s1 > baseCount(i)
          s1=s1-baseCount(i);
          index(numInput-i)=index(numInput-i)+1;
        end;
      end;
      index(numInput)=s1;
      y_center_num=0;
      for i=1:numInput-1
        if index(i) > 1
          y_center_b(s)=y_center_b(s)+y_center(s-baseCount(numInput-i));
          yNum_Fuzzy_Sets(s)=yNum_Fuzzy_Sets(s)+ym(s-baseCount(numInput-i));
          y_center_num=y_center_num+sign(ym(s-baseCount(numInput-i)));
        end;
        if index(i) < fnCounts(i)
          y_center_b(s)=y_center_b(s)+y_center(s+baseCount(numInput-i));
          yNum_Fuzzy_Sets(s)=yNum_Fuzzy_Sets(s)+ym(s+baseCount(numInput-i));
          y_center_num=y_center_num+sign(ym(s+baseCount(numInput-i)));
        end;
      end;
      if index(numInput) > 1
        y_center_b(s)=y_center_b(s)+y_center(s-1);
        yNum_Fuzzy_Sets(s)=yNum_Fuzzy_Sets(s)+ym(s-1);
        y_center_num=y_center_num+sign(ym(s-1));
      end;
      if index(numInput) < fnCounts(numInput)
        y_center_b(s)=y_center_b(s)+y_center(s+1);
        yNum_Fuzzy_Sets(s)=yNum_Fuzzy_Sets(s)+ym(s+1);
        y_center_num=y_center_num+sign(ym(s+1));
      end;
      if y_center_num >= 1
        y_center_b(s)=y_center_b(s)/y_center_num;
        yNum_Fuzzy_Sets(s)=yNum_Fuzzy_Sets(s)/y_center_num;
      end;
    end; % endif ym
  end; % endfor s
  for s=1:numCells
    if ym(s) == 0 & yNum_Fuzzy_Sets(s) ~= 0
      y_center(s)=y_center_b(s);
      ym(s)=yNum_Fuzzy_Sets(s);
    end;
  end;
end; % endwhile ct
end

function yy=eval_wm_deep(Num_Fuzzy_Sets,y_center,ranges,data_matrix)
% Compute the output of WM fuzzy system with y_center and ranges for input data_matrix.
% Original function is here:  https://ieeexplore.ieee.org/ielx7/91/9130783/8788632/code_training_algorithm.pdf?arnumber=8788632

exapp=data_matrix;
[numSamples,m]=size(exapp);
numInput=m;
for i=1:numInput
  fnCounts(i)=Num_Fuzzy_Sets;
end;
activFns = zeros(numInput,2);
activGrades = zeros(numInput,2);
baseCount(1)=1;
for i=2:numInput
  baseCount(i)=1;
for j=2:i
baseCount(i)=baseCount(i)*fnCounts(numInput-j+2);
end;
end;
for j=1:numInput
 for i1=1:2^j
  for i2=1:2^(numInput-j)
   ma(i2+(i1-1)*2^(numInput-j),j)=mod(i1-1,2)+1;
  end;
 end;
end;
e1sum=0;
for k = 1:numSamples
 for i = 1:numInput
  numFns = fnCounts(i);
  nthActive = 1;
  for nthFn = 1:numFns
   grade = Membership_at_x(numFns,nthFn,exapp(k,i),ranges(i,1),ranges(i,2));
   if grade > 0
    activFns(i,nthActive) = nthFn;
    activGrades(i,nthActive) = grade;
    nthActive = nthActive + 1;
   end;
  end;
 end;
 for i=1:numInput
  nn(i,1)=activFns(i,1);
  nn(i,2)=nn(i,1)+1;
  if nn(i,1) == fnCounts(i)
   nn(i,2)=nn(i,1);
  end;
 end;
 a=0;
 b=0;
 for i=1:2^numInput
  indexcell=1;
  grade=1;
  for j=1:numInput
   grade=grade*activGrades(j,ma(i,j));
   indexcell=indexcell+(nn(numInput-j+1,ma(i,numInput-j+1))-1)*baseCount(j);
  end;
  a=a+y_center(indexcell)*grade;
  b=b+grade;
 end;
 yy(k)= a/b; % the fuzzy system output
end; % endfor k
end

function y=Membership_at_x(n,i,x,xmin,xmax)
% Compute the value of the i’th membership function in Fig. 2 at x.
% Original function is here:  https://ieeexplore.ieee.org/ielx7/91/9130783/8788632/code_training_algorithm.pdf?arnumber=8788632

h=(xmax-xmin)/(n-1);

if i==1
 if x < xmin
  y=1;
 end;
 if x >= xmin & x < xmin+h
  y=(xmin-x+h)/h;
 end;
 if x >= xmin+h
  y=0;
 end;
end;

if i > 1 & i < n
 if x < xmin+(i-2)*h | x > xmin+i*h
  y=0;
 end;
 if x >= xmin+(i-2)*h & x < xmin+(i-1)*h
  y=(x-xmin-(i-2)*h)/h;
 end;
 if x >= xmin+(i-1)*h & x <= xmin+i*h
  y=(-x+xmin+i*h)/h;
 end;
end;

if i==n
 if x < xmax-h
  y=0;
 end;
 if x >= xmax-h & x < xmax
  y=(-xmax+x+h)/h;
 end;
 if x >= xmax
  y=1;
 end;
end;
end